# -*- coding: utf-8 -*-
"""Processeurs d'objets ERA
Un processeur est chargé de transformer un objet ERA en règles iptables
"""

from era.backend.iptrule import Target, Mark, TargettedRule, ParameterFactory, DestIPParameter, LogParameter, TCPProtocol
from era.backend.utils import gen_couples
from era.noyau.constants import ACTION_DENY, ACTION_ALLOW
from era.noyau.pool import library_store
try:
    import os
    from era.noyau.qos import QoS
    USE_QOS = True
    # FIXME: Store qos xml filename somewhere
    QOSXML_FILENAME = "qos.xml"
    assert os.path.isfile(QOSXML_FILENAME)
except:
    USE_QOS = False


class DirectiveProcessor(object):
    """Classe de base pour traiter les directives ERA
    """

    def __init__(self, directive):
        """
         - directive : la directive ERA
        """
        self.directive = directive
        # self.directive = self._is_list(directive)
        self.table_name = 'filter'
        self.implicit_rules = []


    def process_time(self, rule):
        """process des paramètres de temps
        """
        if self.directive.time != None:
            time_params = ParameterFactory.time_from_directive(self.directive)
            rule.add_parameter(time_params)
        return rule

    def process_log(self, chain_name, service_param, param_list, match_params=None):
        """Transforme une directive LOG en règles iptables
        """
        # définition du libellé de log
        zone1 = self.directive.get_source_zone()
        zone2 = self.directive.get_dest_zone()
        libelle = "era %s-%s" % (zone1.name[:3],zone2.name[:3])
        param = LogParameter(libelle)
        target = param
        # On crée la règle avec la bonne chaine
        # on met toujours 'filter pour les règles de log' !!
        rule = TargettedRule(target, chain_name, 'filter')
        # mstate
        if library_store.get_version() > 1.0:
            if match_params != None:
                """ajout de la fonctionnalité de -m state (c'est soumis au versionning)"""
                rule.add_parameter(match_params)
        # On ajoute tous les paramètres à la règle
        rule.add_parameter(service_param)
        for param in param_list:
            rule.add_parameter(param)
        # on process les paramètre du module time
        self.process_time(rule)
        return rule

    def _is_list(self, directive):
        """
        Repère si la directive est composée d'une liste d'extemités
        ou d'un groupe de services.
        """
        if directive.src_inv:
            # la source est inversée
            if len(directive.src_list) >= 1 :
                # mais la source est une liste
                return self.invert_src(directive)

        if directive.dest_inv:
            # la destination est à inverser
            if len(directive.dest_list) >= 1:
                # mais la destination est une liste
                return self.invert_dest(directive)

        if directive.serv_inv:
            # le service est à inverser
            if directive.is_group:
                # mais le service est un groupe de services
                return self.invert_serv(directive)
        else:
            # tout va bien
            return directive

    def create_rule(self, target, chain_name, table, match_params, service_param, param_list):
        # On crée la règle avec la bonne chaine
        rule = TargettedRule(target, chain_name, table)
        # mstate
        if library_store.get_version() > 1.0:
            """ajout de la fonctionnalité de -m state (c'est soumis au versionning)"""
            if match_params != None:
                rule.add_parameter(match_params)
        # On ajoute tous les paramètres à la règle
        rule.add_parameter(service_param)
        for param in param_list:
            rule.add_parameter(param)
        # on process les paramètre du module time
        self.process_time(rule)
        return rule

    def get_implicit_rules(self):
        return self.implicit_rules

class BasicDirectiveProcessor(DirectiveProcessor):
    """Processeur d'objet 'Directive' (de type Basic)
    """

    def _inverted(self):
        # traiter self.directive
        pass


    def _basic_inversion(self):
        """inversion de l'action
        Inversion une authorisation en interdiction et réciproquement
        """
        if action == ACTION_ALLOW:
            return ACTION_DENY
        if action == ACTION_DENY:
            return ACTION_ALLOW
        # normalement ça ne passe pas par là
        # puisque c'est une directive basique...
        else:
            return action

    def invert_src(self, directive):
        """Inversion de la source
        """
        # inversion de l'action
        # cela permet de traiter la liste des extremites
        directive.action = self._invert_action(directive.action)
        directive.src_inv = False
        # FIXME -- DROP
        return directive

    def invert_dest(self, directive):
        """Inversion de la destination
        """
        return directive

    def invert_serv(self, directive):
        """Inversion du service
        """
        return directive

    def process(self):
        """Transforme une directive basique en règles iptables
        """
        rules = []
        target = ParameterFactory.target_from_directive(self.directive)
        # FIXME : à dégager
        if USE_QOS and (self.directive.qos_in or self.directive.qos_out):
            has_qos = True
            starqos = QoS(QOSXML_FILENAME)
        else:
            has_qos = False

        # On parcourt les couples d'extrémités utilisées dans la directive
        for extr1, extr2 in gen_couples(self.directive.src_list,
                                        self.directive.dest_list):
            # On récupère l'ensemble des paramètres IPs / Interfaces
            param_set = ParameterFactory.parameters_for_extremites(extr1, extr2,
                                                                   self.directive.src_inv,
                                                                   self.directive.dest_inv)
            chain_name = '%s-%s' % (extr1.zone.name[:3], extr2.zone.name[:3])
            for param_list in param_set:
                # on initialise la liste des paramètres de service
                service_params = ParameterFactory.service_from_directive(self.directive)
                match_params = ParameterFactory.match_state(['NEW'])
                for service_param in service_params:
                    if has_qos:
                        # Create mark
                        mark = 0
                        if self.directive.qos_in:
                            rule = starqos.getRule("in", self.directive)
                            mark |= rule.get_mark()
                        if self.directive.qos_out:
                            rule = starqos.getRule("out", self.directive)
                            mark |= rule.get_mark()
                        # iptables -A mangle -j MARK --or-mark (...)
                        rule = self.create_rule(Mark(mark, "or"), chain_name, "mangle", match_params, service_param, param_list)
                        rules.append(rule)
                        # XXX FIXME : voir avec haypo
                        # iptables -A mangle -j ACCEPT (...)
                        # rule = self.create_rule(Target("ACCEPT"), chain_name, "mangle", match_params, service_param, param_list)
                        # rules.append(rule)

                    # ajout des règles de log
                    if self.directive.is_logged():
                        rule = self.process_log(chain_name, service_param, param_list, match_params)
                        rules.append(rule)

                    # ajout des règles de marquage
                    if self.directive.is_marked():
                        operator, mark = self.directive.mark
                        rule = self.create_rule(Mark(mark, operator), "marquage", "mangle", match_params, service_param, param_list)
                        rules.append(rule)

                    rule = self.create_rule(target, chain_name, self.table_name, match_params, service_param, param_list)
                    rules.append(rule)

        return rules

    def get_implicit_rules(self):
        return []

class DNATDirectiveProcessor(DirectiveProcessor):
    """Processeur d'objet 'Directive' (de type DNAT)
    """

    def __init__(self, directive):
        super(DNATDirectiveProcessor, self).__init__(directive)

    def process(self):
        """Transforme une directive NAT en règles iptables
        """
        rules = []
        target = ParameterFactory.target_from_directive(self.directive)

        # On parcourt les couples d'extrémités utilisées dans la directive
        for extr1, extr2 in gen_couples(self.directive.src_list,
                                        self.directive.dest_list):
            # On récupère l'ensemble des paramètres IPs / Interfaces
            param_set = ParameterFactory.parameters_for_extremites(extr1, extr2,
                                                                   self.directive.src_inv,
                                                                   self.directive.dest_inv)
            main_chain = "PREROUTING"
            for param_list in param_set:
                service_params = ParameterFactory.service_from_directive(self.directive)
                for service_param in service_params:
                    # On crée la règle avec la bonne chaine
                    rule = TargettedRule(target, main_chain, 'nat')
                    # On ajoute tous les paramètres à la règle
                    rule.add_parameter(service_param)
                    in_int, out_int, src_param, dest_param = param_list
                    # PREROUTING => On enlève les '-o'
                    rule.add_parameters((in_int, src_param, dest_param))
                    # paramètres horaires
                    self.process_time(rule)
                    # On ajoute la règle aux autres
                    rules.append(rule)

                for service in self.directive.service:
                    ## Règle implicite : accept de la 1ère vers la 3ème extrêmité
                    extr3 = self.directive.nat_extr
                    chain_name = '%s-%s' % (extr1.zone.name[:3],
                                            extr3.zone.name[:3])
                    service_param = ParameterFactory.service(
                        service.protocol,
                        self.directive.nat_port)
                    # dest_param n'est pas bon (pas celui du comportement par défaut)
                    out_int = ParameterFactory.interface(extr3.zone.interface, True)
                    dest_param = DestIPParameter(extr3.ip_list[0])
                    # ajout des règles de log
                    if self.directive.is_logged():
                        rule = self.process_log(chain_name, service_param, (in_int, out_int, src_param, dest_param))
                        rules.append(rule)
                    implicit_rule = TargettedRule('accept', chain_name)
                    implicit_rule.add_parameter(service_param)
                    # dest_param = DestIPParameter(self.directive.nat_extr.ip_list[0])
                    for param in (in_int, out_int, src_param, dest_param):
                        implicit_rule.add_parameter(param)
                    # paramètres horaires
                    self.process_time(implicit_rule)
                    # ce n'est pas nécessaire si c'est un flux descendant
                    # if self.directive.is_montante():
                    # rules.append(implicit_rule)
                    self.implicit_rules.append(implicit_rule)

                    # ajout des règles de marquage
                    if self.directive.is_marked():
                        operator, mark = self.directive.mark
                        rule = self.create_rule(Mark(mark, operator), "marquage", "mangle", None, service_param, (in_int, out_int, src_param, dest_param))
                        rules.append(rule)

        return rules

    def get_implicit_rules(self):
        return self.implicit_rules


class SNATDirectiveProcessor(DirectiveProcessor):
    """Processeur d'objet 'Directive' (de type SNAT)
    """

    def __init__(self, directive):
       super(SNATDirectiveProcessor, self).__init__(directive)

    def process(self):
        """Transforme une directive NAT en règles iptables
        """
        rules = []
        target = ParameterFactory.target_from_directive(self.directive)

        # On parcourt les couples d'extrémités utilisées dans la directive
        for extr1, extr2 in gen_couples(self.directive.src_list,
                                        self.directive.dest_list):
            # On récupère l'ensemble des paramètres IPs / Interfaces
            param_set = ParameterFactory.parameters_for_extremites(extr1, extr2,
                                                                   self.directive.src_inv,
                                                                   self.directive.dest_inv)
            main_chain = "POSTROUTING"
            for param_list in param_set:
                service_params = ParameterFactory.service_from_directive(self.directive)
                for service_param in service_params:
                    in_int, out_int, src_param, dest_param = param_list
                    # On crée la règle avec la bonne chaine
                    rule = TargettedRule(target, main_chain, 'nat')
                    # On ajoute tous les paramètres à la règle
                    rule.add_parameter(service_param)
                    # POSTROUTING => On enlève les '-i'
                    rule.add_parameters((out_int, src_param, dest_param))
                    # paramètres horaires
                    self.process_time(rule)
                    # On ajoute la règle aux autres
                    rules.append(rule)

                    ## Règle implicite : accept de la 1ère vers la 3ème extrêmité
                    extr3 = self.directive.nat_extr
                    chain_name = '%s-%s' % (extr1.zone.name[:3],
                                            extr2.zone.name[:3])
                    # ajout des règles de log
                    if self.directive.is_logged():
                        rule = self.process_log(chain_name, service_param, param_list)
                        rules.append(rule)
                    implicit_rule = TargettedRule('accept', chain_name)
                    #params = ParameterFactory.parameters_for_zones(extr1.zone,
                    #                                               extr2.zone)
                    implicit_rule.add_parameter(service_param)
                    # implicit_rule.add_parameter((out_int, src_param, dest_param))
                    # XXX FIXME : Demander à Sam s'il ne faut pas mettre un
                    # nat_port pour le service (comme pour le DNAT)
                    for param in param_list:
                        implicit_rule.add_parameter(param)

                    # paramètres horaires
                    self.process_time(implicit_rule)

                    # ce n'est pas nécessaire si la directive est montante...
                    # if self.directive.is_montante():
                    # les regles implicites seront ajoutee ulterieurement
                    # (a la fin du flux)
                    self.implicit_rules.append(implicit_rule)

                    # ajout des règles de marquage
                    if self.directive.is_marked():
                        operator, mark = self.directive.mark
                        rule = self.create_rule(Mark(mark, operator), "marquage", "mangle", None, service_param, param_list)
                        rules.append(rule)
        return rules

    def get_implicit_rules(self):
        return self.implicit_rules


class RedirectDirectiveProcessor(BasicDirectiveProcessor):
    """Processeur pour cible REDIRECT
    Le comportement est le-même que celui par défaut, mais la table
    est 'nat' et pas 'filter
    """
    def __init__(self, directive):
        """
         - directive : la directive ERA
        """
        super(RedirectDirectiveProcessor, self).__init__(directive)
        self.table_name = 'nat'
        self.implicit_rules = []

    def process(self):
        """Transforme une directive basique en règles iptables
        """
        rules = []
        target = ParameterFactory.target_from_directive(self.directive)
        # on récupère le port de redirection pour la règle implicite
        redirect_port = target.dport[0]
        # On parcourt les couples d'extrémités utilisées dans la directive
        for extr1, extr2 in gen_couples(self.directive.src_list,
                                        self.directive.dest_list):
            # On récupère l'ensemble des paramètres IPs / Interfaces
            param_set = ParameterFactory.parameters_for_extremites(extr1, extr2,
                                                                   self.directive.src_inv,
                                                                   self.directive.dest_inv)
            chain_name = 'PREROUTING'
            trigram_zones = '%s-%s' % (extr1.zone.name[:3], extr2.zone.name[:3])
            for in_int, out_int, src_param, dest_param in param_set:
                service_params = ParameterFactory.service_from_directive(self.directive)
                for service_param in service_params :
                    # On crée la règle avec la bonne chaine
                    rule = TargettedRule(target, chain_name, self.table_name)
                    # On ajoute tous les paramètres à la règle
                    rule.add_parameter(service_param)
                    rule.add_parameters([in_int, src_param, dest_param])

                    if library_store.get_version() > 1.0:
                        # la règle n'est pas nécessaire pour les flux descendant
                        # on ajoute la règle implicite uniquement pour des version updatées des modèles

                        # ajout des règles de log
                        implicit_chain = '%s-bas' % extr1.zone.name[:3]
                        dport_param = TCPProtocol(dport = redirect_port)
                        if self.directive.is_logged():
                            rule = self.process_log(implicit_chain, dport_param, param_set)
                            rules.append(rule)
                        # c'est une règle de type input qui est générée
                        implicit_rule = TargettedRule('accept', implicit_chain)
                        # implicit_rule.add_parameter((in_int, src_param, dest_param))
                        implicit_rule.add_parameter(in_int)
                        # service_param c'est le --dport de la règle normale
                        # implicit_rule.add_parameter(service_param)
                        # on ajoute le bon --dport
                        implicit_rule.add_parameter(dport_param)
                        for param in param_set:
                            implicit_rule.add_parameter(param)

                        # paramètres horaires
                        self.process_time(implicit_rule)

                        # paramètres horaires
                        self.process_time(rule)
                        rules.append(implicit_rule)

                    rules.append(rule)
                    # ajout des règles de marquage
                    if self.directive.is_marked():
                        operator, mark = self.directive.mark
                        rule = self.create_rule(Mark(mark, operator), "marquage", "mangle", None, dport_param, param_set)
                        rules.append(rule)

        return rules

    def get_implicit_rules(self):
        return self.implicit_rules


class MARKDirectiveProcessor(DirectiveProcessor):
    """Processeur d'objet 'Directive' (de type MARK)
    """

    def process(self):
        """Transforme une directive MARK en règles iptables
        """
        pass


def get_processor(directive):
    """Fabrique de processeurs
    """
    # on crée les règles de log en premier
    #if directive.is_marked():
    #    yield MARKDirectiveProcessor(directive)
    if directive.action in (1, 2): # ACTION_DENY, ACTION_ALLOW
        yield BasicDirectiveProcessor(directive)
    if directive.action == 4: # ACTION_FORW
        yield RedirectDirectiveProcessor(directive)
    if directive.action == 8: # ACTION_DNAT
        yield DNATDirectiveProcessor(directive)
    if directive.action == 16: # ACTION_MASK
        yield SNATDirectiveProcessor(directive)

