# -*- coding: UTF-8 -*-

# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Module gérant les objets de bases manipulés par l'éditeur
Ces objets sont :
    - les flux
    - les zones
    - les extrémités
    - les services / groupes de services
    - les directives
"""

from itertools import chain

from era.compat import any, sorted, set
from era.noyau.constants import *
from era.noyau.dpatterns import Observer, Observable
from era.noyau.erreurs import BadPortError, TagError
from era.noyau.pool import library_store

# visiteurs pour la sortie rml
try:
    from era.rml import visitors
except:
    # reportlab non présent
    pass

# Classe DirectiveStore ########################################################
class DirectiveStore(Observable):
    """ Modèle pour le classeur de directives
    """

    def __init__(self, directive_list = []):

        Observable.__init__(self)
        self.classeur_directives = []

        for directive in directive_list:
            self.add_directive(directive)


    def add_directive(self, directive, load = 0):
        """ Ajoute une directive au classeur
        """
        if load == 0:
            directive.set_priority(len(self.classeur_directives)+1)
        self.classeur_directives.append(directive)

        self.notify_observers()

    def del_directive(self,index):
        """ Supprime une directive du classeur
        """
        for directive in self.classeur_directives[index+1:]:
            directive.set_priority(directive.priority - 1)
        del(self.classeur_directives[index])
        self.classeur_directives.sort()
        self.notify_observers()

    def swap(self, index1, index2):
        """Echange l'ordre des deux directives pointées par index1 et index2
        """
        self.classeur_directives[index2].set_priority(index1+1)
        self.classeur_directives[index1].set_priority(index2+1)
        self.classeur_directives.sort()
        self.notify_observers()

    def directive_list(self):
        """ renvoie la liste des directives de ce modèle
        """
        return self.classeur_directives


class FwObject:
    """Classe de base des objets du Firewall (directives, extremites, etc.)
    Cette classe définit l'ensemble des caractères communs de tous les objets
    du Firewall. En particulier, chaque objet peut-être 'hérité' d'un modèle,
    et être ou non 'éditable'.
    """

    instanciable_attributes = ()

    def __init__(self, inherited = False, editable = True, mime_type = None):
        """
        inherited : définit si l'objet est hérité d'un modèle ou pas
        editable : définit si l'objet est modifiable ou pas (lié à l'import)
        mime_type : le mime_type de l'objet
        """
        self.inherited = inherited
        self.editable = editable
        self.mime_type = mime_type

    def is_inherited(self):
        """Pour savoir si cet objet a été hérité d'un modèle
        """
        return self.inherited

    def set_inherited(self, inherited):
        """Définit si l'objet est hérité d'un modèle ou pas
        """
        self.inherited = inherited

    def is_editable(self):
        """Renvoie True si l'objet est éditable, False sinon
        """
        return self.editable

    def set_editable(self, editable):
        """Fixe l'éditablilité de l'objet
        """
        self.editable = editable

    def get_type(self):
        """Renvoie le type de l'objet
        """
        return self.mime_type

    def needs_instanciation(self):
        """renvoie True si l'objet contient des variables à instancier,
        sinon False
        """
        return any([getattr(self, attrname, '').startswith('%%')
                    for attrname in self.instanciable_attributes])

    def unspecified_vars(self):
        """renvoie la liste des variables non spécifiés (non instanciées)
        trouvées sur l'objet
        """
        unspecified = set()
        for attrname in self.instanciable_attributes:
            value = getattr(self, attrname, '')
            if value.startswith('%%'):
                if value.endswith('%%'):
                    unspecified.add(value[2:-2])
                else:
                    unspecified.add(value[2:])
        return unspecified


# EXTREMITE #############################################################
class Extremite(FwObject):
    """ Classe représentant une extrémité
    """
    instanciable_attributes = ('name', 'netmask')

    # XXX FIXME : C'est moche le coup du all_zone, trouver mieux !!
    def __init__(self, zone, nom, libelle, ip_list, netmask,
                 all_zone = False, subnet=0):
        """
        zone : la zone dans laquelle est incluse l'extrémité
        nom : le nom de l'extrémité.
        libelle : un descriptif de l'extrémité
        ip_list : la liste des IPs
        netmask : le netmask de l'extrémité
        subnet : indique que l'extremite est un sous-réseau (permet d'utiliser le bon dialogue)
        all_zone : un booléen qui indique l'extrémité est l'ensemble de la zone
        """
        FwObject.__init__(self, False, True, "firewall/extremite")
        self.all_zone = all_zone

        # Si l'extrémité représente toute la zone, elle ne doit pas
        # être éditable
        if self.all_zone:
            self.set_editable(False)
        self.zone = zone
        ## XXX FIXME : pour unifier, changer nom en name
        if all_zone:
            self.name = self.zone.name
        else:
            self.name = nom
        self.libelle = libelle
        self.ip_list = ip_list
        self.netmask = netmask
        self.subnet = subnet
        self.used = 0


    def __str__(self):
        l = ['Nom\t: %s'%self.name]
        l.append(u'Zone\t: %s'%self.zone.name)
        l.append(u'Libellé\t: %s'%self.libelle)
        l.append(u'Netmask\t: %s'%self.netmask)

        return '\n'.join(l)

    def __repr__(self):
        return self.name
        # return "<extremite:%s> %s" % (self.name,id(self))


    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde une extrémité au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<extremite zone="%s" name="%s" libelle="%s" netmask="%s" subnet="%d">\n'% \
                     (" "*indent, self.zone.name, self.name, self.libelle, self.netmask, self.subnet))
        for ip in self.ip_list:
            stream.write('%s<ip address="%s"/>\n'%(" "*(indent+4),ip))
        stream.write('%s</extremite>\n'%(" "*indent))

    def dump_rml(self):
        """Sortie rml d'une extremité
        """
        return visitors.extremite(self.name, self.ip_list, self.netmask)

    def __iter__(self):
        """Pour faciliter l'itération sur les ips
        """
        return iter(self.ip_list)

    def __cmp__(self, other):
        """
        Permet d'ordonner les extremitées dans l'ordre alphabétique
        """
        if other is None:
            return 1
        return cmp(self.name, other.name)

    def needs_instanciation(self):
        """renvoie True si l'objet contient des variables à instancier,
        sinon False
        """
        if FwObject.needs_instanciation(self):
            return True
        return any([ip for ip in self.ip_list if ip.startswith('%%')])


    def unspecified_vars(self):
        """renvoie la liste des variables non spécifiés (non instanciées)
        trouvées sur l'objet
        """
        unspecified = FwObject.unspecified_vars(self)
        for ip in self.ip_list:
            if ip.startswith('%%'):
                if ip.endswith('%%'):
                    unspecified.add(ip[2:-2])
                else:
                    unspecified.add(ip[2:])
        return unspecified

    # XXX FIXME, bouger cette méthode (comme fonction dans ihm/gtk_models ?)
    # On ne veut plus de traces de GTK dans le module noyau.
#    def as_gtk_list_store():
#        """Retourne la représentation en forme de list_store
#        d'une extrémité
#        """
#        store = gtk.ListStore(gobject.TYPE_STRING,  # name
#                              gobject.TYPE_STRING, # libelle
#                              gobject.TYPE_STRING) # zone.name
#
#        return store
#
#    as_gtk_list_store = staticmethod(as_gtk_list_store)

class Directive(FwObject):
    """
    Classe définissant un ensemble de règles
    """

    def __init__(self, src_list, dest_list, service, action, attrs,
                 nat_extr=None, nat_port=None, is_group=False,
                 src_inv=False, dest_inv=False, serv_inv=False,
                 libelle="pas de description", time=None,
                 user_group=None, mark=None, qos_in=None, qos_out=None):
        """
        src_list : la liste des extrémités sources
        dest_list : la liste des extrémités destinations
        service : le service (ou groupe) associé à la directive
        action : un masque définissant les actions
        attrs : une valeur définissant si la règle est optionnelle et /ou active
        src_inv : flag d'inversion de l'extremité source
        dest_inv : flag d'inversion de l'extremité destination
        serv_inv : flag d'inversion du service
        libelle : un descriptif de la directive (et donc du groupe de règles)
        user_group : groupe d'utilisateurs (peut être None)
        mark : marquage des paquets [opérateur, marque]
        qos_in: rêgle de QoS en entrée, identifiant numérique (peut être None)
        qos_out: rêgle de QoS en sortie, identifiant numérique (peut être None)
        """
        FwObject.__init__(self, False, True, "firewall/directive")
        assert service is not None, _("service undefined error")
        self.libelle = libelle
        self.time = time
        self.src_list = src_list
        self.src_inv = src_inv
        self.dest_list = dest_list
        self.dest_inv = dest_inv
        self.service = service
        self.serv_inv = serv_inv
        self.action = action
        self.attrs = int(attrs)
        self.priority = 0
        self.nat_extr = None
        self.nat_port = None
        self.mark = mark
        self.service_group = is_group
        self.user_group = user_group
        self.qos_in = qos_in
        self.qos_out = qos_out
        if not self.action_is_default():
            if not self.action == 4 :
                assert nat_extr is not None, _("nat extr undefined error")

            assert nat_port is not None, _("nat port undefined error")
            self.nat_extr = nat_extr

            try:
                self.nat_port = int(nat_port)
            except ValueError:
                raise BadPortError, _("bad port error")

            # vérifions la validitié du port
            # README : le port 0 signifie TOUT (tous les ports)
            if not 0 <= self.nat_port <= 65535 and not self.nat_port=="TOUT":
                raise BadPortError, _("bad port error")

    def action_is_default(self):
        """Retourne vrai si l'action est DENY ou ALLOW, false sinon
        """
        return self.action <= 2

    def is_active(self):
        """renvoie 'Vrai' si la règle est active
        """
        # FIXME : regarder dans le library_store pour voir si la directive est active
        return self.attrs & DIRECTIVE_ACTIVE == DIRECTIVE_ACTIVE

    def is_optional(self):
        """ renvoie 'Vrai' si la règle est optionnelle
        """
        return self.attrs & DIRECTIVE_OPTIONAL == DIRECTIVE_OPTIONAL


    def is_logged(self):
        """ renvoie 'Vrai' si on doit journaliser les évènements relatifs à
        cette directive
        """
        return self.attrs & DIRECTIVE_LOG == DIRECTIVE_LOG


    def is_marked(self):
        """ renvoie 'Vrai' si on doit marquer cette règle
        """
        return self.mark is not None


    def is_hidden(self):
        """ renvoie 'Vrai' si on doit marquer cette règle
        """
        return self.attrs & DIRECTIVE_HIDDEN == DIRECTIVE_HIDDEN


    def get_source_zone(self):
        """ Renvoie la zone source de la directive
        """
        return self.src_list[0].zone


    def get_dest_zone(self):
        """ Renvoie la zone destination de la directive
        """
        return self.dest_list[0].zone


    def is_montante(self):
        """ renvoie 'Vrai' si la directive part d'une zone moins sécurisée vers
        une zone plus sécurisée
        """
        return self.get_source_zone() < self.get_dest_zone()


    def get_directive_type(self):
        """Renvoie le type de la directive (UP ou DOWN)
        Méthode crée par convenance
        """
        src_zone = self.get_source_zone()
        dest_zone = self.get_dest_zone()
        if src_zone <= dest_zone:
            return DIRECTIVE_TYPE_UP
        else:
            return DIRECTIVE_TYPE_DOWN


    def set_priority(self, priority):
        """Définit la priorité de la directive
        priority : un entier définissant la priorité
        """
        self.priority = priority

    def set_time(self, time):
        """Définit les options horaires de la directive
        priority : un dictionnaire avec les bonnes options
        """
        self.time = time


#    def update_attrs(self, col_number):
#
#        if col_number == LOG_COLUMN:
#            self.attrs = self.attrs ^ DIRECTIVE_LOG
#        elif col_number == MARK_COLUMN:
#            self.attrs = self.attrs ^ DIRECTIVE_MARK
#        elif col_number == ACTIVE_COLUMN:
#            self.attrs = self.attrs ^ DIRECTIVE_ACTIVE
#        elif col_number == OPTIONAL_COLUMN:
#            self.attrs = self.attrs ^ DIRECTIVE_OPTIONAL
#        else:
#            print "Warning : type d'action inconnu"

    def action_to_string(self):
        d = {ACTION_DENY : _('deny'),
             ACTION_ALLOW : _('allow'),
             ACTION_FORWARD : _('forward'),
             ACTION_DNAT : _('dnat'),
             ACTION_MASK : _('mask')}

        return d[self.action]


    def __cmp__(self, other):
        """ Permet de comparer la priorité de deux directives
        dir1 < dir2 si dir1.priority < dir2.priority
        """
        if other is None:
            return 1
        return cmp(self.priority, other.priority)


    def __str__(self):
        l = ['Directive : ']
        l.append('\t src = %s'%self.src_list)
        l.append('\t dest = %s'%self.dest_list)
        l.append('\t service = %s'%self.service)
        l.append('\t priority = %s'%self.priority)
        l.append('\t action = %s'%self.action_to_string())
        l.append('\t log = %s'%self.is_logged())
        l.append('\t mark = %s'%self.is_marked())
        l.append('\t active = %s'%self.is_active())
        l.append('\t optional = %s'%self.is_optional())

        return '\n'.join(l)

    def set_tag(self, tag):
        """Affectation d'un attribut de tag la directive
        si celle-ci est optionnelle.
        """
        if self.is_optional():
            self.tag = tag
            if tag == "":
                raise TagError, _("empty_tag")
        else :
            raise TagError

    def get_tag(self):
        """Renvoie le tag de la directive s'il existe.
        """
        if hasattr(self, "tag"):
            return self.tag
        else :
            raise TagError

    def set_attrs(self, attrs):
        """active ou désactive une directive optionnelle
        """
        self.attrs = int(attrs)

    def get_attrs(self):
        """renvoie l'attrs
        """
        return self.attrs

    def get_libelle(self):
        """renvoie le libelle
        """
        return self.libelle

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde une directive au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<directive ' % (" "*indent, ) )
        if self.is_optional():
            try:
                stream.write('tag="%s" ' % self.tag)
            except:
                raise TagError, "une directive optionnelle n'a pas de tag"

        # si la directive est optionnelle, on regarde dans library_store.tags si le tag est actif ou caché
        try:
            if self.is_optional():
                if library_store.tags[str(self.tag)][0] == True:
                    self.attrs = self.attrs | DIRECTIVE_ACTIVE
                else:
                    self.attrs = self.attrs & 29
                if library_store.tags[str(self.tag)][1] == True:
                    self.attrs = self.attrs | DIRECTIVE_HIDDEN
                else:
                    self.attrs = self.attrs & 15
        except Exception, e:
            print "erreur : ",e

        stream.write('service="%s" priority="%d" action="%d" attrs="%d" ' % \
                     (self.service.name, self.priority, self.action, self.attrs)
                     )

        if self.user_group != None:
            stream.write('user_group="%s" ' % self.user_group.id)
        if self.mark != None:
            stream.write('mark_operator="%s" mark_value="%s" ' % (self.mark[0], self.mark[1]))

        if self.action > 2:
            if self.action != 4:
                # cas de SNAT et DNAT
                stream.write('nat_extr="%s" nat_port="%d" '% (self.nat_extr.name, self.nat_port ))
            else:
                # cas de la redirection
                stream.write('nat_port="%d" '% (self.nat_port))

        stream.write('src_inv="%d" dest_inv="%d" serv_inv="%d" libelle="%s" '% \
                         ( self.src_inv, self.dest_inv, self.serv_inv, self.libelle))
        if self.time != None:
            stream.write('time="%s">\n' % self.time.name)
        else:
            stream.write('>\n')


        for src in self.src_list:
            stream.write('%s<source name="%s"/>\n'%(" "*(indent+4),src.name))
        for dest in self.dest_list:
            stream.write('%s<destination name="%s"/>\n'%(" "*(indent+4),dest.name))
        stream.write('%s</directive>\n'%(" "*indent))

    def dump_rml(self):
        """Exportation en rml des directives
        """
        return visitors.directive(self._rml_list())

    def _rml_list(self):
        """ Renvoie une liste représentant la directive.
        pour le formatage en rml.
        """
        # si les extremites sont inversées, on affiche un "!" devant
        if self.src_inv == 1:
            src_list = "!"+str(self.src_list)
        else:
            src_list = str(self.src_list)

        if self.dest_inv == 1:
            dest_list = "!"+str(self.dest_list)
        else:
            dest_list = str(self.dest_list)

        if self.serv_inv == 1:
            service = "!" + self.service.libelle
        else:
            service = self.service.libelle

        return [self.priority,src_list, dest_list, self.action_to_string(), service, self.libelle]

    def as_list(self):
        """ Renvoie une liste représentant la directive.
        L'élément N de la liste renvoyée est la valeur de l'attribut se trouvant
        à la colonne N du modèle de l'arbre GTK
        """
        # si les extremites sont inversées, on affiche un "!" devant
        if self.src_inv == 1:
            src_list = "!"+str(self.src_list)
        else:
            src_list = str(self.src_list)

        if self.dest_inv == 1:
            dest_list = "!"+str(self.dest_list)
        else:
            dest_list = str(self.dest_list)

        if self.serv_inv == 1:
            service = "!" + str(self.service)
        else:
            service = str(self.service)

        if hasattr(self.time, 'name'):
            timerange = self.time.name
        else:
            timerange = _("undefined")

        if hasattr(self.user_group, 'name'):
            user_group = self.user_group.name
        else:
            user_group = _("undefined")

        if self.mark is not None:
            # operation (set, and, or) et marque
            mark = "%s (%s)" % (self.mark[1],self.mark[0])
        else:
            mark = _("undefined")

        return [self.priority,src_list, dest_list, service, self.action_to_string(),timerange,user_group,mark,self.is_logged(),
                self.is_optional(), self]

    def needs_instanciation(self):
        """renvoie True si l'objet contient des variables à instancier,
        sinon False
        """
        if self.nat_extr:
            checks = (self.src_list, self.dest_list, self.service, [self.nat_extr])
        else:
            checks = (self.src_list, self.dest_list, self.service)
        return any([obj.needs_instanciation() for obj in chain(*checks)])

    def unspecified_vars(self):
        """renvoie la liste des variables non spécifiés (non instanciées)
        trouvées sur l'objet
        """
        if self.nat_extr:
            checks = (self.src_list, self.dest_list, self.service, [self.nat_extr])
        else:
            checks = (self.src_list, self.dest_list, self.service)
        return set(chain(*[obj.unspecified_vars() for obj in chain(*checks)]))


# CLASSE SERVICE ###############################################################
class Service(FwObject):
    instanciable_attributes = ('protocol',)

    def __init__(self, name, protocol, port_list, id, libelle) :
        """
        name : the service's name
        protocol : the concerned protocol
        port_list : a list of all concerned ports
        id : l'identifiant du service
        libelle : description du service
        """
        FwObject.__init__(self, False, True, "firewall/service")
        self.name = name
        self.protocol = protocol
        self.port_list = port_list
        self.id = id
        self.libelle = libelle
        self.used = 0

    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux services ayant même nom sont égaux.
        """
        if other is None:
            return False
        return self.name == other.name


    def __cmp__(self, other):
        """ Permet de trier alphabétiquement deux services
        """
        if other is None:
            return 1
        return cmp(self.name, other.name)


    def dump_rml(self):
        """
        renvoie la sortie rml d'un service.
        """
        return visitors.service(self)

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde un service au format interne
        stream : le flux dans lequel on écrit
        """
        port_str = self.port_list[0]
        if len(self.port_list) > 1:
            port_str = "%s-%s"%(self.port_list[0], self.port_list[-1])
        stream.write('%s<service name="%s" protocol="%s" ports="%s" id="%s" libelle="%s" />\n'% \
                     (" "*indent,self.name, self.protocol, port_str, self.id, self.libelle))

    def __iter__(self):
        """XXX **Bof** : pour ne pas avoir à différencier un service d'un groupe
        """
        return iter([self])


    def __str__(self):
        return "service : %s, protocol = %s"%(self.name, self.protocol)


    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)


# CLASSE GROUPESERVICE #########################################################
class ServiceGroupe(FwObject):
    """Groupe de service
    """
    def __init__(self, id, libelle, services=[]) :
        """
        id : le nom (qui sert aussi d'identifiant) du groupe
        services : liste des services du groupe
        libelle : le descriptif du groupe
        """
        FwObject.__init__(self, False, True, "firewall/group")
        self.id = id
        self.services = services
        self.libelle = libelle
        self.name = self.id
        self.used = 0

    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux services ayant même nom sont égaux.
        """
        if other is None:
            return False
        return self.id == other.id


    def __cmp__(self, other):
        """ Permet de trier alphabétiquement deux services
        """
        if other is None:
            return 1
        return cmp(self.id, other.id)


    def add_service(self, service) :
        """Ajoute un service au groupe de services
        et vérifie que le service n'est pas déjà existant
        """
        if (service in self.services) :
            raise ExistingServiceError, _('existing service')
        else :
            self.services.append(service)

    def __str__(self):
        liste_serv = []
        longueur = len(self.services)
        for gr in self.services[0:3]:
            liste_serv.append(gr.name)
        return "groupe %s : %s"%(self.id,liste_serv)

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde un groupe au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<groupe id="%s" libelle="%s">\n'%(" "*indent, self.id, self.libelle))
        for service in self.services:
            service.save(stream, indent+4)
        stream.write('%s</groupe>\n'%(" "*indent))

    def __iter__(self):
        """Pour simplifier l'itération sur les services du groupe de services"""
        return iter(self.services)

    def dump_rml(self):
        return ''.join([s.dump_rml() for s in self.services])

    def needs_instanciation(self):
        """renvoie True si l'objet contient des variables à instancier,
        sinon False
        """
        return any([s.needs_instanciation() for s in self.services])

    def unspecified_vars(self):
        """renvoie la liste des variables non spécifiés (non instanciées)
        trouvées sur l'objet
        """
        return set(chain(*[s.unspecified_vars() for s in self.services]))

class UserGroup(FwObject):
    """Utilisateur ou groupe"""
    def __init__(self, id, name):
        """
        id : entier (gid du groupe)
        name : le descriptif du groupe
        """
        FwObject.__init__(self, False, True, "firewall/user_group")
        self.id = id
        self.name = name
        # used : indique si l'objet est référencé dans un autre objet
        self.used = 0

    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux services ayant même nom sont égaux.
        """
        if other is None:
            return False
        return self.name == other.name

    def __cmp__(self, other):
        """ Permet de trier alphabétiquement deux services
        """
        if other is None:
            return 1
        return cmp(self.name, other.name)

    def __str__(self):
        return "<user_group:%s>" % self.name

    def __repr__(self):
        return "<user_group:%s> %s" % (self.name,id(self))

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde une zone au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<user_group name="%s" id="%s"/>\n'% (" "*indent, self.name, self.id))

class Zone(FwObject):
    """ Classe Zone
    """
    instanciable_attributes = ('ip', 'network', 'netmask', 'interface')

    def __init__(self, name, level, ip, network="0.0.0.0",
                 netmask="255.255.255.255", interface="lo"):
        """name : nom de la zone"""
        FwObject.__init__(self, False, True, "firewall/zone")
        self.name = name
        self.level = int(level)
        self.ip = ip
        self.network = network
        self.netmask = netmask
        self.interface = interface
        self._listeners = []

    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux zones ayant même nom (et même niveau)
        sont égales. Sinon, avec __cmp__, deux zones de noms différents, mais
        de niveaux identiques seraient considérées comme égales.
        """
        if other is None:
            return False

        return self.name == other.name and self.level == other.level

    def strict_eq(self, other):
        if other is None:
            return False
        return self.level == other.level


    def __cmp__(self, other):
        """ Permet de comparer le niveau de deux zones
        zone1 < zone2 si zone1.level < zone2.level
        """
        if other is None:
            return 1
        if self.level == other.level:
            return cmp(self.name,other.name)
        else:
            return cmp(self.level, other.level)

    def __str__(self):
        return "<zone:%s>" % self.name

    def __repr__(self):
        return "<zone:%s> %s" % (self.name,id(self))


    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde une zone au format interne
        stream : le flux dans lequel on écrit
        """
#        if self.shares:
#            stream.write('%s<zone name="%s" level="%d" ip="%s" network="%s" netmask="%s" interface="%s">\n'% \
#                     (" "*indent, self.name, self.level, self.ip, self.network, self.netmask, self.interface))
#            for name, percent in self.shares.items():
#                stream.write('%s <band name="%s" percent="%s"/>\n' % ("  "*indent, name, str(percent)))
#            stream.write('</zone>\n')
#        else:
        stream.write('%s<zone name="%s" level="%d" ip="%s" network="%s" netmask="%s" interface="%s"/>\n'% \
                     (" "*indent, self.name, self.level, self.ip, self.network, self.netmask, self.interface))


    def dump_rml(self):
        return visitors.zone(self.name, self.level)

    def add_listener(self, listener):
        """ Ajouter un auditeur de l'objet zone
        """
        # assert listener not in self._listeners
        self._listeners.append(listener)


    def notify_listeners(self):
        """ Méthode qui prévient tous les auditeurs d'un changement
        """
        for l in self._listeners:
            l.zone_changed()



# CLASSE FLUX ##################################################################
class Flux(FwObject,Observer):

    def __init__(self, zoneA, zoneB):
        """
        Attention : le flux n'est pas orienté ! Il n'y a pas de différence entre :
        f = Flux(z1,z2) et f = Flux(z2,z1)
        zoneA et zoneB : les deux extrémités du flux
        """
        FwObject.__init__(self, False, True, "firewall/flux")

        self.zoneA = zoneA
        self.zoneB = zoneB

        # self.directives = []
        self.up_directives_store = DirectiveStore()
        self.down_directives_store = DirectiveStore()
        self.up_directives_store.register_observer(self)
        self.down_directives_store.register_observer(self)

        self._listeners = []


    def __eq__(self, other):
        """Retourne vrai si les deux flux s'appliquent aux mêmes zones
        """
        if other is None:
            return 0
        return sorted([other.zoneA.name, other.zoneB.name]) == \
               sorted([self.zoneA.name, self.zoneB.name])


    def up_directives_model(self):
        """ Renvoie le modèle associé aux directives montantes
        """
        return self.up_directives_store
        # return [d for d in self.directives if d.is_montante()]


    def down_directives_model(self):
        """ Renvoie le modèle associé aux directives descendantes
        """
        return self.down_directives_store
        # return [d for d in self.directives if not d.is_montante()]


    def up_directives_list(self):
        """ Renvoie la liste des directives montantes de ce flux
        Cette méthode est un raccourci pour :
        f.up_directives_model().directive_list()
        """
        return self.up_directives_store.directive_list()


    def down_directives_list(self):
        """ Renvoie la liste des directives descendantes de ce flux
        Cette méthode est un raccourci pour :
        f.down_directives_model().directive_list()
        """
        return self.down_directives_store.directive_list()

    def update(self):
        """Appelé lorsque l'un des directive_store a changé
        """
        self.notify_listeners()

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde un flux au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<flux zoneA="%s" zoneB="%s">\n'%(" "*indent,
                                                         self.zoneA.name, self.zoneB.name))
        stream.write('%s<montantes>\n'%(" "*(indent+4)))

        #si la liste n'est pas vide, on la trie par ordre de priorité
        if len(self.up_directives_list()) > 0:
            self.up_directives_list().sort()

        for directive in self.up_directives_list():
            # XXX FIXME : Trick pour pas sauver en double lors d'héritage
            if not directive.is_inherited():
                directive.save(stream, indent+8)

        stream.write('%s</montantes>\n'%(" "*(indent+4)))
        stream.write('%s<descendantes>\n'%(" "*(indent+4)))

        #si la liste n'est pas vide, on la trie par ordre de priorité
        if self.down_directives_list() > 0:
            self.down_directives_list().sort()

        for directive in self.down_directives_list():
            # XXX FIXME : Trick pour pas sauver en double lors d'héritage
            if not directive.is_inherited():
                directive.save(stream, indent+8)

        stream.write('%s</descendantes>\n'%(" "*(indent+4)))
        stream.write('%s</flux>\n'%(" "*indent))



    def add_listener(self, listener):
        """ Ajouter un auditeur de l'objet flux
        """
        # assert listener not in self._listeners
        self._listeners.append(listener)


    def notify_listeners(self):
        """ Méthode qui prévient tous les auditeurs d'un changement
        """
        # XXX FIXME : on devrait passer le type de changement à la méthode
        # flux_changed() (Par exemple, la directive ajoutée)
        for l in self._listeners:
            l.flux_changed()


    def __str__(self):
        return "Flux : [%s <===> %s]"%(self.zoneA.name, self.zoneB.name)

    def __repr__(self):
        return "Flux : [%s <===> %s]"%(self.zoneA.name, self.zoneB.name)

    def dump_rml(self):
        """Exportation au format XML
        """
        return visitors.flux(self)

    def needs_instanciation(self):
        """renvoie True si l'objet contient des variables à instancier,
        sinon False
        """
        # py2.3
        return any([d.needs_instanciation()
                    for d in chain(self.up_directives_list(), self.down_directives_list())])

    def unspecified_vars(self):
        """renvoie la liste des variables non spécifiés (non instanciées)
        trouvées sur l'objet
        """
        return set(chain(*[d.unspecified_vars()
                           for d in chain(self.up_directives_list(), self.down_directives_list())
                           if d.needs_instanciation()]))



# CLASSE RANGE ###############################################################
class Range(FwObject):

    def __init__(self, range_dict) :
        """
        name : the service's name
        range_dict:  parameters for time module
        """
        FwObject.__init__(self, False, True, "firewall/range")
        self.used = 0
        self.name = range_dict['name']
        for opt in ('timestart','timestop','datestart','datestop','days'):
            setattr(self, opt, range_dict.get(opt, ''))

    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux services ayant même nom sont égaux.
        """
        if other is None:
            return False
        return self.name == other.name

    def __cmp__(self, other):
        """ Permet de trier alphabétiquement deux services
        """
        if other is None:
            return 1
        return cmp(self.name, other.name)


    def dump_rml(self):
        """
        renvoie la sortie rml d'un service.
        """
        return visitors.ranges(self)

    def save(self, stream, indent = 0):
        """Méthode qui sauvegarde un service au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<range name="%s"' % (" "*indent,self.name))
        if self.timestart != '':
            stream.write(' timestart="%s"' % self.timestart)
        if self.timestop != '':
            stream.write(' timestop="%s"' % self.timestop)
        if self.datestart != '':
            stream.write(' datestart="%s"' % self.datestart)
        if self.datestop != '':
            stream.write(' datestop="%s"' % self.datestop)
        if self.days != '':
            stream.write(' days="%s"' % self.days)
        stream.write(' />\n')

    def __str__(self):
        txt= ["%s :" % self.name]
        if self.days != '':
            txt.append(" (%s)" % self.days)
        if self.timestart != '':
            txt.append(" %s - %s" % (self.timestart, self.timestop))
        if self.datestart != '':
            txt.append(" %s - %s" % (self.datestart, self.datestop))
        return ''.join(txt)


# CLASSE DE QOS ###############################################################
class QosClass(FwObject):
    """classe de qos """

    def __init__(self, id, bandwidth, libelle="pas de description", priority=1, zone=None) :
        """
        id : l'identifiant à partir duquel le bitmask pourra être récupéré
        bandwidth : un pourcentage
        zone : le nom de zone à laquelle la qos appartient
        """
        FwObject.__init__(self, False, True, "firewall/qosclass")
        self.id = id
        self.priority = priority
        self.bandwidth = float(bandwidth)
        self.libelle = libelle
        self.zone = zone
        
    def __eq__(self, other):
        """ On redéfinit __eq__ en plus de __cmp__ pour s'assurer
        que seules deux services ayant même nom sont égaux.
        """
        if other is None:
            return False
        return self.id == other.id

    def __cmp__(self, other):
        """ Permet de trier alphabétiquement deux services
        """
        if other is None:
            return 1
        return cmp(self.id, other.id)

    def set_bandwidth(self, bandwidth):
        """permet de remettre le bandwidth en float
        (la bande passante est en string quand elle vient de la fenêtre de répartition)
        """
        self.bandwidth = float(bandwidth)

    def save(self, stream, indent = 0):
        """ Méthode qui sauvegarde une classe de qos au format interne
        stream : le flux dans lequel on écrit
        """
        stream.write('%s<qos id="%s" priority="%s" zone="%s" libelle="%s" bandwidth="%f"/>\n'% \
                     (" "*indent, self.id, self.priority, self.zone, self.libelle, self.bandwidth))

    def __str__(self):
        return "qosclass : %s, bandwidth = %f"%(self.id, self.bandwidth)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.zone)
