# -*- coding: UTF-8 -*-
"""
Module qui permet d'initialiser l'application à partir de fichiers
En particulier, ce module contient les fonctions de parsages "haut-niveau"
(<=> le DOM se fait plus bas), et la fonction initialize_app qui est utilisée
pour (ré)initiliaser l'application (au démarrage, ou dans "Open", ou dans
"Load")

"""

from domparsers import instantiate_zone_from_dom, get_services_from_dom, \
     instantiate_group_from_dom, instantiate_extremite_from_dom, \
     instantiate_flux_from_dom, instantiate_include_from_dom, \
     instantiate_range_from_dom, instantiate_user_group_from_dom, \
     instantiate_qos_from_dom, instanciate_bandwidth_from_dom

from era.noyau.path import *
from era.noyau.constants import *
from xml.dom.minidom import parse
from era.noyau.models import MatrixModel
from era.noyau.pool import library_store
import os

LIB_PATH = os.path.join(ERA_DIR,"lib")
DEFAULT_LIB_PATH = os.path.join(LIB_PATH,"defaults")
ZONES_FILENAME = "zones.xml"
SERVICES_FILENAME = "services.xml"
USER_GROUPS_FILENAME = "user_groups.xml"

def load_defaults(matrix_model):
    """Charge les zones et services par défaut de l'application
    matrix_model : le modèle de matrice de flux dans lequel on ajoute
    les zones
    """
    zone_list = parse_zone_file(os.path.join(DEFAULT_LIB_PATH,ZONES_FILENAME))
    for z in zone_list:
        matrix_model.add_zone(z)
    load_default_services()
    load_default_user_group()

def load_default_services():
    """ Charge la liste des services par défaut et l'ajoute à la liste passée
    en paramètre.
    """
    service_list, group_list = parse_service_file(os.path.join(DEFAULT_LIB_PATH,SERVICES_FILENAME))
    library_store.services = service_list
    library_store.groups = group_list

def add_default_services():
    """Charge la liste des services par défaut et l'ajoute éventuellement
    à la liste des services du library_store si ces services 
    ne sont pas déjà présents
    """
    service_list, group_list = parse_service_file(os.path.join(DEFAULT_LIB_PATH, SERVICES_FILENAME))
    for serv in service_list:
        if serv not in library_store.services:
            library_store.add_service(serv)
    for grp in group_list:
        if grp not in library_store.groups:
            library_store.add_group(grp)

def load_default_user_group():
    """ Charge la liste des user_group par défaut et l'ajoute à la liste passée
    en paramètre.
    """
    library_store.user_groups = parse_user_group_file(os.path.join(DEFAULT_LIB_PATH,USER_GROUPS_FILENAME))

def parse_firewall_file(filepath, inherited = False):
    """Parse un fichier firewall.
    filepath : le path du fichier à parser
    inherited : True si on parse le fichier en tant que modèle, False sinon
    """
    try:
        # On récupère le DOM document, puis le noeud racine
        document = parse(filepath)
        root_node = document.firstChild
        return _parse_firewall_file(root_node, inherited)
    except Exception, e:
#        import traceback
#        traceback.print_exc()
        raise Exception("ATTENTION : Probleme serieux dans le fichier xml source")

def _parse_firewall_file(root_node, inherited):

    # Si ce firewall hérite d'un modèle => on parse ce modèle
    modl = root_node.getAttribute('model')
    if root_node.hasAttribute('model') and str(modl) != "":
        zones, flux_list, extremites, services, groups, model_rules, netbios, ranges, model, user_groups, qos, options = \
               parse_firewall_file(root_node.getAttribute('model'), inherited = True)
        # le modèle retourné sera toujours celui du fichier ouvert (pas les modèles de modèles)
        model = root_node.getAttribute('model')
    else:
        services, groups = {}, {}
        zones = {}
        extremites = {}
        ranges = {}
        flux_list = []
        model = None
        user_groups = {}
        qos = []
        options = {}
        
    ## parsing des options du modèle (netbios, qos en mode utilisateur ou normal)
    options = {}
    if root_node.hasAttribute('netbios'):
        # l'attribut netbios doit etre 0 ou 1
        netbios = int(root_node.getAttribute('netbios'))
    else:
        # si l'attribut netbios n'est pas renseigné
        netbios = 1
    options['netbios'] = netbios
# FIXME : pas de prise en compte de l'attribut user (utilisateur de classe de qos)
#    
#    if root_node.hasAttribute('user'):
#        # l'attribut user doit etre 0 ou 1
#        user = int(root_node.getAttribute('user'))
#    else:
#        # si l'attribut user n'est pas renseigné
#        user = 0
#    options['user'] = user
    
    if root_node.hasAttribute('version'):
        # l'attribut version doit etre 0 ou 1
        version = float(root_node.getAttribute('version'))
        """stockage de la version directement dans le library_store"""
        library_store.set_version(version)
    else:
        # si pas de version --> ancien modèle, on considère qu'on est en version 1.0
        library_store.set_version('1.0')


    services_node = root_node.getElementsByTagName('services')[0]
    sl, gl = get_services_from_dom(services_node, inherited = inherited)
    services = dict([(s.name, s) for s in sl])
    groups = dict([(g.id, g) for g in gl])
    
    # FIXME : prendre en compte les bandwidths
    # bandwidths
    upload, download = instanciate_bandwidth_from_dom(root_node.getElementsByTagName('qosclasses'))
    # qos classes
    qos = instantiate_qos_from_dom(root_node.getElementsByTagName('qos'))
    
    user_group_nodes = root_node.getElementsByTagName('user_group')
    for child in user_group_nodes:
        user_group = instantiate_user_group_from_dom(child, inherited = False)
        user_groups[user_group.id] = user_group

    # On crée les objets plage horaires (range)
    try:
        ranges = instantiate_range_from_dom(root_node, inherited = inherited)
    except:
        ranges = {}

    # On crée les zones
    for zone_elt in root_node.getElementsByTagName('zone'):
        zone = instantiate_zone_from_dom(zone_elt, inherited = inherited)
        zones[zone.name] = zone

    # On crée les extrémités
    for extremite_elt in root_node.getElementsByTagName('extremite'):
        extr = instantiate_extremite_from_dom(extremite_elt, zones, inherited = inherited)
        extremites[extr.name] = extr

    # On crée les flux
    for flux_elt in root_node.getElementsByTagName('flux'):
        flux = instantiate_flux_from_dom(flux_elt,zones, flux_list, extremites, services, groups, ranges, user_groups, inherited = inherited)
        flux_list.append(flux)

    # On crée éventuellement l'inclusion statique
    try:
        rule = root_node.getElementsByTagName('include')[0]
        rules = instantiate_include_from_dom(rule)
    except:
        rules = None

    # FIXME : upload et download sont momentanement places dans options (beurk)
    options = (options, upload, download)

    # On retourne l'ensemble des objets créés
    return zones, flux_list, extremites, services, groups, rules, netbios, ranges, model, user_groups, qos, options

def parse_zone_file(filepath):
    """Fonction qui parse un fichier librairie de zones
    filepath : le path du fichier à parser.
    """
    document = parse(filepath).firstChild
    zone_nodes = document.getElementsByTagName('zone')
    zone_list = []
    for node in zone_nodes:
        zone_list.append(instantiate_zone_from_dom(node))

    return zone_list

def parse_user_group_file(filepath):
    """Fonction qui parse un fichier librairie de services
    filepath : le path du fichier à parser.
    """
    document = parse(filepath).firstChild
    user_groups = {}
    user_group_nodes = document.getElementsByTagName('user_group')
    for child in user_group_nodes:
        user_group = instantiate_user_group_from_dom(child, inherited = False)
        user_groups[user_group.id] = user_group
    return user_groups

def parse_service_file(filepath):
    """Fonction qui parse un fichier librairie de services
    filepath : le path du fichier à parser.
    """
    document = parse(filepath).firstChild
    # services_node = document.getElementsByTagName('services')[0]
    return get_services_from_dom(document)

def create_matrix_model(zones, flux, inherited=None):
    """Crée et initialise un modèle de matrice avec le dictionnaire
    zones et la liste flux
    """
    matrix_model = MatrixModel()
    # On ne passe pas par la méthode add_zone pour ne pas
    # créer les flux automatiquement
    for zone in zones.values():
        matrix_model.add_zone(zone)
    matrix_model.zones.sort()
    
    # On réécrit par dessus les flux créés automatiquement
    # dans add_zone()
    matrix_model.flux = flux
    for f in matrix_model.flux:
        f.add_listener(matrix_model)

    matrix_model.inherited = inherited
        
    return matrix_model


def init_library_store(extremites, services, groups, rules, netbios, ranges, user_groups, qos, options):
    """réinitialise library_store avec les extremites, services et
    groupes passés en paramètre
    options = {'netbios':netbios}
    qos : qosclasses
    """
    library_store.extremites = extremites
    library_store.services = services.values()
    library_store.qos = qos
    library_store.groups = groups.values()
    library_store.rules = rules
    library_store.ranges = ranges
    library_store.user_groups = user_groups
    library_store.set_options(options)
    
def initialize_app(fw_file = None, inherited = False):
    """fonction d'initialisation de l'application
    fw_file : le filepath du fichier firewall. Si fw_file vaut None,
    on retourne des modèles vides
    Retourne le modèle de matrice de flux créé.
    """
    if fw_file is not None:
        zones, flux_list, extremites, services, groups, rules, netbios, ranges, model, user_groups, qos, options = \
               parse_firewall_file(fw_file, inherited)
        matrix_model = create_matrix_model(zones, flux_list, model)
        # FIXME: les bandwiths sont pour l'instant rangees dans les options
        init_library_store(extremites, services, groups, rules, netbios, ranges, user_groups, qos, options[0]) 
        library_store.set_bandwidth(options[1], options[2])
        library_store.clear_tags()
        # recharge les libs par defaut si necessaire
        add_default_services()

        # recupere les tags des directives optionnelles
        try:
            tags = matrix_model.get_active_list()
        except:
            tags = []

        for i in tags:
            # convertissons en 
            active = i[1] & DIRECTIVE_ACTIVE == DIRECTIVE_ACTIVE
            hidden = i[1] & DIRECTIVE_HIDDEN == DIRECTIVE_HIDDEN
            library_store.tags[str(i[0])]=(active,hidden)
    else:
        library_store.clear()
        matrix_model = MatrixModel()
        load_defaults(matrix_model)
    return matrix_model

